/*
 * Copyright (c) 2019 UPLEX Nils Goroll Systemoptimierung
 * All rights reserved
 *
 * Author: Geoffrey Simmons <geoffrey.simmons@uplex.de>
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package varnish

import (
	"time"

	"code.uplex.de/uplex-varnish/varnishapi/pkg/admin"
	"github.com/prometheus/client_golang/prometheus"
)

const (
	// UpdateErr is the Event reason name when a VCL update fails
	UpdateErr    = "UpdateFailure"
	connectErr   = "ConnectFailure"
	vsmErr       = "SharedMemoryFailure"
	pingErr      = "PingFailure"
	statusErr    = "StatusFailure"
	statusNotRun = "StatusNotRunning"
	restartErr   = "RestartFailure"
	restart      = "Restart"
	panicErr     = "PanicFailure"
	panic        = "Panic"
	vclListErr   = "VCLListFailure"
	discardErr   = "VCLDiscardFailure"
	monitorGood  = "MonitorGood"
)

func (vc *Controller) checkInst() bool {
	metrics.monitorChecks.Inc()

	vc.log.Infof("Monitoring Varnish")

	vc.admMtx.Lock()
	defer vc.admMtx.Unlock()
	vc.wg.Add(1)
	defer vc.wg.Done()

	vc.log.Tracef("Get address and secret from %s", vc.home)
	addr, secret, err := vc.getAddrSecret()
	if err != nil {
		metrics.vsmFails.Inc()
		vc.ErrorEvt(vsmErr, "Error reading admin address and secret: "+
			"%v", err)
		return false
	}

	timer := prometheus.NewTimer(metrics.connectLatency)
	adm, err := admin.Dial(addr, secret, vc.admTimeout)
	timer.ObserveDuration()
	if err != nil {
		metrics.connectFails.Inc()
		vc.ErrorEvt(connectErr, "Error connecting: %v", err)
		return false
	}
	defer adm.Close()
	vc.Banner = adm.Banner
	vc.log.Debugf("Connected to Varnish instance, banner: %s", adm.Banner)

	pong, err := adm.Ping()
	if err != nil {
		metrics.pingFails.Inc()
		vc.ErrorEvt(pingErr, "Error pinging: %v", err)
		return false
	}
	metrics.pings.Inc()
	vc.log.Debugf("Succesfully pinged instance: %+v", pong)

	state, err := adm.Status()
	if err != nil {
		vc.ErrorEvt(statusErr, "Error getting status: %v", err)
		return false
	}
	if state == admin.Running {
		metrics.childRunning.Inc()
		vc.log.Debugf("Status: %s", state)
	} else {
		metrics.childNotRunning.Inc()
		vc.WarnEvt(statusNotRun, "Status: %s", state)
		// XXX configure whether to auto-start
		if state == admin.Stopped {
			vc.log.Debug("Attempting restart")
			if err := adm.Start(); err != nil {
				vc.WarnEvt(restartErr, "Cannot restart: %v",
					err)
				metrics.restartFails.Inc()
				return false
			}
			vc.InfoEvt(restart, "Child restarted")
			metrics.restarts.Inc()
		}
	}

	panic, err := adm.GetPanic()
	if err != nil {
		vc.ErrorEvt(panicErr, "Error getting panic: %v", err)
		return false
	}
	if panic == "" {
		vc.log.Debug("No panic")
	} else {
		metrics.panics.Inc()
		vc.ErrorEvt(panic, "Panic: %s", panic)
		// XXX clear the panic? Should be configurable
	}

	vcls, err := adm.VCLList()
	if err != nil {
		vc.ErrorEvt(vclListErr,
			"Error getting VCL list: %v", err)
		return false
	}
	for _, vcl := range vcls {
		if vcl.Temperature == admin.ColdTemp {
			if err = adm.VCLDiscard(vcl.Name); err != nil {
				vc.ErrorEvt(discardErr,
					"Error discarding VCL %s: %v",
					vcl.Name, err)
				return false
			}
			metrics.vclDiscards.Inc()
			vc.log.Infof("Discarded VCL %s", vcl.Name)
		}
	}
	return true
}

func (vc *Controller) monitor(monitorIntvl time.Duration) {
	if monitorIntvl <= 0 {
		vc.log.Infof("Varnish monitor interval=%v, monitor not running",
			monitorIntvl)
		return
	}
	vc.log.Info("Varnish monitor starting, interval: ", monitorIntvl)

	for {
		time.Sleep(monitorIntvl)

		vc.log.Infof("Monitoring Varnish instance")

		good := true
		if !vc.checkInst() {
			good = false
		}

		if vc.updater == nil {
			vc.ErrorEvt(UpdateErr, "Updater not initialized")
			metrics.updateErrs.Inc()
			good = false
		} else if err := vc.updater.Update(); err != nil {
			vc.ErrorEvt(UpdateErr, "Errors updating Varnish: %+v",
				err)
			metrics.updateErrs.Inc()
			good = false
		}
		if good {
			vc.InfoEvt(monitorGood, "Monitor check good")
		}
	}
}
